import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import warnings
import re


def find_function_type_from_header(header):
    pattern = "(.*)function"
    real_type_pattern = r"real\((.*)\)"
    match = re.search(pattern, header.strip(), re.IGNORECASE)
    fun_type = remove_brackets_content(match.groups(0)[0].strip().lower())
    if fun_type == "real":
        real_declaration = match.group(1)
        nt = re.match(real_type_pattern, real_declaration, re.I)
        try:
            fun_type = VariablePrecision.lookup_table[nt.group(1)]
        except AttributeError:
            fun_type = VariablePrecision.wp

    ret_val = None
    if header.count("result"):
        ret_val = header.lower().split('result')[1]
        ret_val = ret_val.replace('results', '')
        ret_val = ret_val.replace('(', '')
        ret_val = ret_val.replace(')', '')
    if fun_type:
        fun_name = header.lower().split('function')[1]
        fun_name = fun_name.split("(")[0]
        fun_name = fun_name.replace("(", "")
        fun_name = fun_name.strip()
        if not ret_val:
            if match.group(0).count('rpe_var'):
                return 'rpe_var', fun_name
            else:
                return fun_type, fun_name
        else:
            if match.group(0).count('rpe_var'):
                return 'rpe_var', ret_val
            else:
                return fun_type, ret_val
    else:
        if header.count("result"):
            pattern = r"result *\((.*)\)"
            match = re.search(pattern, header, re.IGNORECASE)
            if match:
                return None, remove_brackets_content(match.groups(0)[0].strip().lower())
            else:
                raise ValueError("Not finding function result")

        else:
            pattern = r"function (.*)\("
            match = re.search(pattern, header, re.IGNORECASE)
            if match:
                return None, remove_brackets_content(match.groups(0)[0].strip().lower())
            else:
                raise ValueError("Not finding function result")


def find_var_elsewhere(variable, file_objects):
    pattern = r"[^\w]%s[^\w]"

    files = ["%s.f90" % module.name for module in variable.routine.module.used_by]
    files.append("%s.f90" % variable.routine.module.name)
    files_to_search = []
    for f in file_objects:
        if f in files:
            files_to_search.append(file_objects[f])
        if len(files_to_search) == files:
            break

    return [f for f in files_to_search if re.search(pattern % variable.name, f.text)]


def find_file_and_line_of_allocation(variable, file_objects):
    if is_allocated_inside_declaration_scope(variable, file_objects):
        f = file_objects["%s.f90" % variable.routine.module.name]
        for line_index, line in enumerate(f.lines):
            if re.search("allocate.*[^a-zA-Z0-9_%]" + variable.name + "[^a-zA-Z0-9_%]", line, re.IGNORECASE):
                return f, line_index
        raise ValueError("Allocation of variable %s not found" % variable.name)
    else:
        files_where_var_appears = find_var_elsewhere(variable, file_objects)
        for f in files_where_var_appears:
            for line_index, line in enumerate(f.lines):
                if re.search(r"allocate *\(.*[^a-zA-Z0-9_%]" + variable.name + "[^a-zA-Z0-9_%]", line, re.IGNORECASE):
                    return f, line_index
        warnings.warn("Allocation of variable %s not found" % variable.name)
        return None, None


def find_line_in_sourcefile(variable, sourcefile):
    pattern = r"namelist.*[^\w]%s[^\w]" % variable.name
    namelist_name = None
    for line in sourcefile.lines:
        if re.search(pattern, line, re.IGNORECASE):
            pattern2 = r"namelist/(.*)/.*%s[^\w]" % variable.name
            match = re.search(pattern2, line, re.IGNORECASE)
            namelist_name = match.group(1)
            break
    if namelist_name:
        for index, line in enumerate(sourcefile.lines):
            pattern3 = r"read *\(.*[^\w]%s[^\w]" % namelist_name
            if re.search(pattern3, line, re.IGNORECASE):
                return sourcefile, index
    return None, None


def find_file_and_line_of_namelist_reading(variable, file_objects):
    routine = variable.routine
    module = routine.module
    sourcefile = file_objects["%s.f90" % module.name]

    sourcefile, index = find_line_in_sourcefile(variable, sourcefile)

    if sourcefile:
        return sourcefile, index

    other_sources = find_var_elsewhere(variable, file_objects)
    for sourcefile in other_sources:
        sourcefile, index = find_line_in_sourcefile(variable, sourcefile)
        if sourcefile:
            return sourcefile, index

    return None, None


def find_first_use(variable, file_objects):
    routine = variable.routine
    module = routine.module

    sourcefile = file_objects["%s.f90" % module.name]

    # pattern = '(^|\W+)%s(\W+|$)' % variable.name
    pattern = r'\W*%s\W*' % variable.name
    searching = False
    loop_level = 0
    loop_start = None
    in_where_statement = False
    where_statement_start = None

    # Scroll all the lines
    for line_index, line in enumerate(sourcefile.lines):
        line = line.strip().lower()
        # Until it finds the subroutine to which the variables belongs (routine.name)
        if not searching:
            match = re.search(r"(subroutine|function) *(\w*)", line, re.IGNORECASE)
            if match:
                called_routine = match.group(2)
                if called_routine.lower() == routine.name.lower():
                    searching = True
        # In the subroutine
        else:
            # Scroll all the lines searching for the first use
            if re.search(pattern, line, re.IGNORECASE):
                # skip declarations: this is a really bad workaround, but avoids false positive that will block compilation.
                if line.count("::") or line.lower().count("implicit none"):
                    continue
                # Skip the first line
                elif line.count("subroutine"):
                    # Reach the end: the variable has not been used
                    if line.count("end subroutine"):
                        break
                    continue
                # Skip import statements
                elif line.find("use ") == 0:
                    continue
                # Other statement
                elif line.find("data ") == 0:
                    continue
                else:
                    # The variable is found
                    if in_where_statement:
                        # Inside a where statement
                        return sourcefile, where_statement_start
                    else:
                        # In a (nested) loop
                        if loop_level > 0:
                            return sourcefile, loop_start
                        # On a normal line
                        else:
                            return sourcefile, line_index
            elif line.count("end subroutine"):
                break
            # Whenever a 'do loop' starts increment nesting index
            elif re.search(r"^ *do *[\w]* *=", line, re.IGNORECASE):
                if loop_level == 0:
                    loop_start = line_index
                loop_level += 1
            # Whenever a 'do loop' ends decrement nesting index
            # This way match bot the expression 'end do' and 'enddo' but excludes matches like 'enddo loop_name'
            elif re.search(r"^ *end\s*do", line, re.IGNORECASE):
                loop_level -= 1
            # Do the same for where statement
            elif line.find("where") == 0:
                in_where_statement = True
                where_statement_start = line_index
            elif line.find("end where") == 0:
                in_where_statement = False
    return None, None


def find_routine_text(variable, file_objects):
    routine = variable.routine
    module = routine.module

    f = file_objects["%s.f90" % module.name]
    to_return = []
    lines = f.lines
    if routine.name == "main":
        for line in lines[:]:
            if line.lower().count("contains"):
                break
            else:
                to_return.append(line)
        to_return.append(lines[-1])
    else:
        printing = False
        for line in lines:
            if not printing:
                if line.lower().count("subroutine"):
                    called_routine = re.search(r"subroutine *(\w*)", line, re.IGNORECASE).group(1)
                    if called_routine.lower() == routine.name.lower():
                        to_return.append(line)
                        printing = True
            else:
                to_return.append(line)
                if line.lower().count("end subroutine"):
                    break

    return to_return


## PRIVATE
def is_allocated_inside_declaration_scope(var, file_objects):
    text = find_routine_text(var, file_objects)
    interesting_lines = [line for line in text if
                         re.search(r"allocate *\(.*[^\w]%s[^\w]" % var.name, line, re.IGNORECASE)]
    if interesting_lines:
        return True
    else:
        return False


def remove_brackets_content(string):
    op = [i for i, x in enumerate(string) if x == "("]
    cp = [i for i, x in enumerate(string) if x == ")"]
    _new_line = []
    for n in range(len(string)):
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x <= n]) == len([x for x in cp if x < n]):
            _new_line += [string[n]]
    return "".join(_new_line)
